// onto2viz.js version 1.0
// (C) 2011, KMI, Open University
// Petr Kremen, petr.kremen@gmail.com

// do not use _FAKE_2  and _FAKE_1 as node names
var defaultParams =  {
    "canvasBG"                  : "#FFFFFF",
    "nodeLabelColor"            : "#238ece",
    "nodeColor"                 : "#FFFFC0",
    "nodeStrokeColor"           : "#fcddbe",
    "nodeStrokeExternalColor"   : "#C4D4FF",
    "nodeExternalColor"         : "#ECF4FF",
    "nodeColorRoot"             : "#FFDB87",
    "nodeStrokeColorRoot"       : "#FF9E85",
    "nodeLabelSize"             : 10,
    "edgeLabelValidColor"       : "green",
    "edgeLabelColor"            : "black",
    "edgeColor"                 : "#ff7f00",
    "edgeInferredColor"         : "lightgreen",
    "edgeExternalColor"         : "#C4D4FF",
    "edgeLabelSize"             : 10,
    "labelAlongEdge"            : true,
    "arrowAlpha"                : 2*Math.PI/5,
    "arrowRadius"               : 15,          
    "textScaleFactor"           : 1, // TODO rename
    "repulsion"                 : 1000,
    "stiffness"                 : 2,
    "friction"                  : 0,
    "rho"                       : 0.1
}

var Onto2Viz = function(id, par) {    
    onto2viz = this;
    
    // default values
    params = defaultParams;
        
    // override default values
    for (p in par) {
        params[p] = par[p];
    }
    
    // arbor particle system
    var sys = arbor.ParticleSystem(params.repulsion, params.stiffness, params.friction); // create the system with sensible repulsion/stiffness/friction
    sys.parameters({
        gravity:true
    }) // use center-gravity to make the graph settle nicely (ymmv)
    var canvas = document.getElementById(id);
    var ctx = canvas.getContext("2d");
    var particleSystem = null;
        
    // vars
    this.parsys = sys;
    
    // 'click on node' callback
    this.clickCB = null;

    // 'is selected' callback
    this.isSelectedCB= null;

    // 'is external' callback
    this.isExternalCB= null;
      
    that = {
        init:function(system){
            particleSystem = system;
            particleSystem.screen({
                size:{
                    width:canvas.width, 
                    height:canvas.height
                },
                padding:[36,60,36,60]
            })
            //            $(window).resize(that.resize)

            that.initMouseHandling();
        },

        redraw:function(){
            ctx.fillStyle = params.canvasBG;
            ctx.fillRect(0,0, canvas.width, canvas.height);

            ctx.font="bold "+params.edgeLabelSize+"px sans-serif";
            
            particleSystem.eachEdge(function(edge, p1, p2){
                if ( edge.data.length ==0) {
                    return;
                }
                
                var l = 0;
                
                for( e in edge.data ) {
                    if (edge.data[e].invisible) {
                        continue;
                    }
                    
                    if (edge.data[e].stmttype=='inferred') {
                        ctx.strokeStyle = params.edgeInferredColor;
                        if(edge.data[e].weight != null) {                        
                            ctx.lineWidth = edge.data[e].weight * 10;
                        } else {
                            ctx.lineWidth = 2;
                        }
                    } else if (edge.data[e].stmttype=='external') {
                        ctx.strokeStyle = params.edgeExternalColor;
                        if(edge.data[e].weight != null) {                        
                            ctx.lineWidth = edge.data[e].weight * 10;
                        } else {
                            ctx.lineWidth = 1;
                        }
                    } else {
                        ctx.strokeStyle = params.edgeColor;
                        if(edge.data[e].weight != null) {                        
                            ctx.lineWidth = edge.data[e].weight * 10;
                        } else {
                            ctx.lineWidth = 2;
                        }
                    }
                    ctx.fillStyle = ctx.strokeStyle;
                    
                    var result = {
                        // angle of the connector to p2.
                        eta:null, 
                        // end point of the link (where the link meets the edge of the node)
                        pk : null,
                        // point where to draw text
                        ph: null
                    };                           
            
                    wSource = ctx.measureText(getEffectiveText(ctx,edge.source.data.label)).width;
                    wTarget = ctx.measureText(getEffectiveText(ctx,edge.target.data.label)).width;
           
                    if ( sys.getEdges(edge.source,edge.target).length + sys.getEdges(edge.target,edge.source).length == 1 && edge.target!=edge.source && edge.data.length == 1) {
                        result = drawLine(ctx,p1, p2, wSource, wTarget);
                    } else {
                        if ( edge.source == edge.target ) {
                            result = drawLink(ctx,p1, p2, 1 + (0.1*l),wTarget);                
                        } else {
                            result = drawLink(ctx,p1, p2,params.rho*(1+3*l),wTarget);                
                        }
                    }                
            
                    if (edge.data[e].oriented) {            
                        drawArrowEnding(ctx,result.pk,result.eta);
                    }

                    yShift = -5;//params.edgeLabelSize;                

                    if ( params.labelAlongEdge ) {
                        ctx.save();
                        ctx.translate(result.ph.x,result.ph.y);
                        if (p2.x > p1.x) {
                            ctx.rotate(Math.atan2(p2.y-p1.y,p2.x- p1.x));
                        } else {
                            ctx.rotate(Math.atan2(p2.y-p1.y,p2.x- p1.x) + Math.PI);
                        }
                        ctx.fillText(edge.data[e].label, -ctx.measureText(edge.data[e].label).width/2, yShift); 
                        
                        ctx.restore();                        
                    } else {
                        ctx.fillText(edge.data[e].label, result.ph.x-ctx.measureText(edge.data[e].label).width/2,result.ph.y+yShift);
                    }
                    
                    l++;
                }
            })

            ctx.font="bold "+params.nodeLabelSize+"px sans-serif";
            
            particleSystem.eachNode(function(node, pt){
                if ( node.data.invisible ) {
                    return;
                }
                
                l = node.data.label.split('\n');
    
                w = 0;
    
                for( x in l ) {
                    max = ctx.measureText(getEffectiveText(ctx,l[x])).width;
                    if ( max > w ) {
                        w = max;
                    }
                }
                                
                if ( onto2viz.isSelectedCB != null &&  onto2viz.isSelectedCB( node ) ) {                
                    ctx.strokeStyle = params.nodeStrokeColorRoot;
                    ctx.lineWidth = 2;
                    ctx.fillStyle = params.nodeColorRoot;
                } else if (onto2viz.isExternalCB != null &&  onto2viz.isExternalCB( node ) ) {
                    ctx.lineWidth = 1;
                    ctx.strokeStyle = params.nodeStrokeExternalColor;
                    ctx.fillStyle = params.nodeExternalColor;                    
                } else {
                    ctx.lineWidth = 1;
                    ctx.strokeStyle = params.nodeStrokeColor;
                    ctx.fillStyle = params.nodeColor;
                }             
                ctx.beginPath();
                ctx.arc(pt.x, pt.y, w/2+2*params.textScaleFactor, 0, Math.PI*2, false);
                ctx.closePath();
                ctx.fill();
                ctx.stroke();

                ctx.fillStyle = params.nodeLabelColor;                
                
                for( x in l ) {
                    ctx.fillText(getEffectiveText(ctx,l[x]), pt.x-w/2, pt.y-(l.length/2-x-0.5)*params.nodeLabelSize);
                }

            })    			
        },
      
        initMouseHandling:function(){
            var dragged = null;
            var handler = {
                clicked:function(e){
                    var pos = $(canvas).offset();
                    _mouseP = arbor.Point(e.pageX-pos.left, e.pageY-pos.top);
                    dragged = particleSystem.nearest(_mouseP);

                    if (dragged && dragged.node !== null){                       
                        dragged.node.fixed = true;
                    }

                    $(canvas).bind('mousemove', handler.dragged);
                    $(window).bind('mouseup', handler.dropped);

                    return false;
                },
                dragged:function(e){
                    var pos = $(canvas).offset();
                    var s = arbor.Point(e.pageX-pos.left, e.pageY-pos.top);

                    if (dragged && dragged.node !== null){
                        var p = particleSystem.fromScreen(s);
                        dragged.node.p = p;
                        dragged.realDrag = true;
                    }

                    return false;
                },

                dropped:function(e){
                    if (dragged===null || dragged.node===undefined) return;
                    if (dragged.node !== null) {
                        if(!dragged.realDrag && onto2viz.clickCB != null) {                            
                            onto2viz.clickCB(dragged.node);
                        }
                        dragged.node.fixed = false;
                    }
                    dragged.node.tempMass = 1000;
                    dragged = null;
                    $(canvas).unbind('mousemove', handler.dragged);
                    $(window).unbind('mouseup', handler.dropped);
                    _mouseP = null;
                    return false;
                }
            }       
            // start listening
            $(canvas).mousedown(handler.clicked);
        }
    }
    
    sys.renderer=that;       
}

Onto2Viz.prototype.setData = function(graphdata) {
    //            if (graphdata !== undefined &&  jQuery.isEmptyObject(graphdata.edges)) {
    if ( graphdata === undefined ) {
        graphdata={
            nodes:{},
            edges:{}
        };
    }
    
    graphdata.nodes._FAKE_1 = {
        "invisible" : true
    };
    graphdata.nodes._FAKE_2 = {
        "invisible" : true
    };
    graphdata.edges._FAKE_1= {
        _FAKE_2 : [{
            "invisible":true
        }]
    }
    //            } ;   // hot fix, TODO: make the edge as invisible as possible             
    
    this.parsys.merge(graphdata);
}


Onto2Viz.prototype.addData = function(graphdata) {
    this.parsys.graft(graphdata);
}

Onto2Viz.prototype.getEdgesFrom = function(node) {
    return this.parsys.getEdgesFrom(node);
}

Onto2Viz.prototype.pruneEdgesOfType = function(type, nodeNamesToPreserve) {
    sys = this.parsys;
    this.parsys.eachEdge(function(edge, p1, p2) {
        for( e in edge.data ) {
            if ( edge.data[e].stmttype==type) {
                sys.pruneEdge(edge);
            }
        }
    });
    
    this.parsys.prune(function(node, e) {
        return ($.inArray(node.name,nodeNamesToPreserve) < 0) && (e.from.length + e.to.length == 0);        
    });
}

Onto2Viz.prototype.pruneEdgesOfName = function(name, nodeNamesToPreserve) {
    sys = this.parsys;
    this.parsys.eachEdge(function(edge, p1, p2) {
        for( e in edge.data ) {
            if ( edge.data[e].id == name) {
                sys.pruneEdge(edge);
            }
        }
    });
    
    this.parsys.prune(function(node, e) {
        return ($.inArray(node.name,nodeNamesToPreserve) < 0) && (e.from.length + e.to.length == 0);        
    });
}

Onto2Viz.prototype.setOnClickHandler = function(handler) {
    this.clickCB = handler;
}

Onto2Viz.prototype.setIsSelectedHandler = function(handler) {
    this.isSelectedCB = handler;
}

Onto2Viz.prototype.setIsExternalHandler = function(handler) {
    this.isExternalCB = handler;
}

/**
* Draws an arrow ending at the specified point with the given base angle
*/ 
drawArrowEnding = function(ctx, pk, angle) {
    ctx.beginPath();            
    
    xh = Math.sin(angle-params.arrowAlpha)*params.arrowRadius;
    yh = Math.cos(angle-params.arrowAlpha)*params.arrowRadius;                                
    ctx.moveTo(pk.x,pk.y);                
    ctx.lineTo(pk.x-xh,pk.y-yh);
            
    yh = Math.sin((-Math.PI/2-angle)-params.arrowAlpha)*params.arrowRadius;
    xh = Math.cos((-Math.PI/2-angle)-params.arrowAlpha)*params.arrowRadius;                                
    ctx.moveTo(pk.x-xh,pk.y-yh);                
    ctx.lineTo(pk.x,pk.y);
    
    ctx.stroke();
    ctx.closePath();
}

getEffectiveText = function(ctx,txt) {
    shortened= false
    while( (ctx.measureText(txt).width) > 100 ) {
        shortened =true;
        txt = txt.substring(0,txt.length-2);
    }

    if ( shortened ) {
        txt +="...";
    }  
    return txt;
}

/**
* Draws an arc between two points p1, p2, given the ratio zeta between the height and radius and the width of the source and target label text
*/ 
drawLink = function(ctx,p1, p2, zeta, wTarget) {
    ctx.beginPath();
    
    gamma = Math.atan2(p2.y-p1.y,p2.x-p1.x);

    // loop
    if ( p1.x== p2.x && p1.y == p2.y ) {
        r = zeta*wTarget/2;
        pc = {
            x:p1.x, 
            y:p1.y-r
        }
        ctx.arc(pc.x,pc.y,r,0,2*Math.PI,false);
    } else {
        alpha = 2*Math.asin(Math.sqrt(zeta/2));
        r = Math.sqrt(Math.pow((p1.x-p2.x),2) + Math.pow((p1.y-p2.y),2))/(2*Math.sin(alpha));        
        pc = {
            x:(p1.x+p2.x)/2+r*(1-zeta)*Math.cos(Math.PI/2+gamma), 
            y:(p1.y+p2.y)/2+r*(1-zeta)*Math.sin(Math.PI/2+gamma)
        }
        ctx.arc(pc.x,pc.y,r,-alpha+gamma-Math.PI/2,alpha+gamma-Math.PI/2,false);
    }

    // draw arrow
    w = wTarget/2;
    ksi = Math.asin(w/(2*r));
    //    ksi = 0;//Math.asin(w/(2*r));
    eta = 3*Math.PI/2+2*ksi-Math.atan2(pc.y-p2.y,pc.x-p2.x);
                
    ctx.stroke();
    ctx.closePath();    

    return {
        eta: eta, 
        pk:{
            x:(pc.x+r*Math.sin(eta)), 
            y:(pc.y+r*Math.cos(eta))
        }, 
        ph:{
            x:pc.x-r*Math.cos(Math.PI/2+gamma), 
            y:pc.y-r*Math.sin(Math.PI/2+gamma)
        }
    }
}

/**
* Draws a line between two points p1, p2, given the width of the label text
*/
drawLine = function(ctx, p1, p2, wSource, wTarget) {
    ctx.beginPath();

    var beta = Math.atan2(p2.x-p1.x,p2.y-p1.y);                
    wS = wSource/2;
    wT = wTarget/2;
    pS= {
        x:p1.x + Math.sin(beta)*wS, 
        y:p1.y + Math.cos(beta)*wS
    }

    pT= {
        x:p2.x - Math.sin(beta)*wT, 
        y:p2.y - Math.cos(beta)*wT
    }

    ctx.moveTo(pS.x, pS.y);                
    ctx.lineTo(pT.x, pT.y);
    ctx.stroke();
    ctx.closePath();   

    return {
        eta: beta+Math.PI/2, 
        pk:pT,
        ph : {
            x:(pS.x+pT.x)/2, 
            y:(pS.y+pT.y)/2
        }
    };
}
